// Copyright (c) 2009-2013 Intel Corporation
// All rights reserved.
//
// WARRANTY DISCLAIMER
//
// THESE MATERIALS ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THESE
// MATERIALS, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Intel Corporation is the author of the Materials, and requests that all
// problem reports or change requests be submitted to it directly

__kernel void gemm_naive(__global const float * restrict A,
                         __global const float * restrict B,
                         __global float * restrict C,
						 float alpha, float beta, int width0, int width1)
{
	int row = get_global_id(1);
	int col = get_global_id(0);
	float sum = 0;
	for (int i = 0; i < width0; i++) 
		sum += A[row*width0+i] * B[i*width1+col];

	C[row*width1+col] = alpha * sum + beta * C[row*width1+col];
}

////////////////////////////////////////////////////////////////
// L3_SLM_Fast_16x16_2
#define VEC_SIZE        4
#define TILE_M          128
#define TILE_K          64
#define TILE_N          128
#define ROWS_PER_WI     8

__attribute__((reqd_work_group_size(16, 16, 1)))
__kernel void L3_SLM_8x8_16x16(
    const __global float4 *src0,
    const __global float4 *src1,
    __global float4 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    // Src0 atile is M rows x K columns
    // M = 128
    // K = 64 = 16 float4s
    // requires sizeof(float) x M x K = 32K SLM
    __local float4 atile[TILE_M * TILE_K / VEC_SIZE];

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
    const int local_y = get_local_id(1);

    // Result ctile is M rows x N columns
    // M = 128, we have 16 rows of work-items, so we need 128/16 8 results down
    // N = 128, we have 16 columns of work-items, so we need 128/16 = 8 results across = 2 float4s across

    float4 dot00 = (float4)(0.f);
    float4 dot01 = (float4)(0.f);
    float4 dot02 = (float4)(0.f);
    float4 dot03 = (float4)(0.f);
    float4 dot04 = (float4)(0.f);
    float4 dot05 = (float4)(0.f);
    float4 dot06 = (float4)(0.f);
    float4 dot07 = (float4)(0.f);

    float4 dot10 = (float4)(0.f);
    float4 dot11 = (float4)(0.f);
    float4 dot12 = (float4)(0.f);
    float4 dot13 = (float4)(0.f);
    float4 dot14 = (float4)(0.f);
    float4 dot15 = (float4)(0.f);
    float4 dot16 = (float4)(0.f);
    float4 dot17 = (float4)(0.f);

    __global float4 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( ( group_y * TILE_M ) + ROWS_PER_WI * local_y ) * width1;

    // Src0 is used to load atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    const __global float4 *src0_read = src0 + local_x + ( ( group_y * TILE_M ) + ROWS_PER_WI * local_y ) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    // K = 64, we'll process four rows at a time
    // N = 128, 16 work-items are participating, so we need 128/16 = 8 floats, which is 2 float4s
    const __global float4 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );
    const __global float4 *src1_read1 = src1_read0 + ( TILE_N / 2 / VEC_SIZE );

    __local float4 *slm = atile + local_y * ( ROWS_PER_WI * TILE_K / VEC_SIZE );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        // We want to load atile, which is M rows x K columns
        // M = 128, and we have 16 rows of work-items, so each work-item must load 128/16 = 8 rows.
        // K = 64, and we have 16 columns of work-items, so each work-item must load 64/16 = 4 columns = 1 float4.
        slm[local_x + 0 * TILE_K / VEC_SIZE] = src0_read[0 * width0];
        slm[local_x + 1 * TILE_K / VEC_SIZE] = src0_read[1 * width0];
        slm[local_x + 2 * TILE_K / VEC_SIZE] = src0_read[2 * width0];
        slm[local_x + 3 * TILE_K / VEC_SIZE] = src0_read[3 * width0];
        slm[local_x + 4 * TILE_K / VEC_SIZE] = src0_read[4 * width0];
        slm[local_x + 5 * TILE_K / VEC_SIZE] = src0_read[5 * width0];
        slm[local_x + 6 * TILE_K / VEC_SIZE] = src0_read[6 * width0];
        slm[local_x + 7 * TILE_K / VEC_SIZE] = src0_read[7 * width0];

        src0_read += TILE_K / VEC_SIZE;

        barrier(CLK_LOCAL_MEM_FENCE);

        int i = 0;
        do
	{
         //We get better performance by loading btile first.
            const float4 brow00 = src1_read0[0];   src1_read0 += width1;
            const float4 brow01 = src1_read0[0];   src1_read0 += width1;
            const float4 brow02 = src1_read0[0];   src1_read0 += width1;
            const float4 brow03 = src1_read0[0];   src1_read0 += width1;
            const float4 brow10 = src1_read1[0];   src1_read1 += width1;
            const float4 brow11 = src1_read1[0];   src1_read1 += width1;
            const float4 brow12 = src1_read1[0];   src1_read1 += width1;
            const float4 brow13 = src1_read1[0];   src1_read1 += width1;

            const float4 a0 = slm[ i + 0 * TILE_K / VEC_SIZE ];
            dot00 = mad(brow00, (float4) a0.x, dot00);
            dot00 = mad(brow01, (float4) a0.y, dot00);
            dot00 = mad(brow02, (float4) a0.z, dot00);
            dot00 = mad(brow03, (float4) a0.w, dot00);
            dot10 = mad(brow10, (float4) a0.x, dot10);
            dot10 = mad(brow11, (float4) a0.y, dot10);
            dot10 = mad(brow12, (float4) a0.z, dot10);
            dot10 = mad(brow13, (float4) a0.w, dot10);

            const float4 a1 = slm[ i + 1 * TILE_K / VEC_SIZE ];
            dot01 = mad(brow00, (float4) a1.x, dot01);
            dot01 = mad(brow01, (float4) a1.y, dot01);
            dot01 = mad(brow02, (float4) a1.z, dot01);
            dot01 = mad(brow03, (float4) a1.w, dot01);
            dot11 = mad(brow10, (float4) a1.x, dot11);
            dot11 = mad(brow11, (float4) a1.y, dot11);
            dot11 = mad(brow12, (float4) a1.z, dot11);
            dot11 = mad(brow13, (float4) a1.w, dot11);

            const float4 a2 = slm[ i + 2 * TILE_K / VEC_SIZE ];
            dot02 = mad(brow00, (float4) a2.x, dot02);
            dot02 = mad(brow01, (float4) a2.y, dot02);
            dot02 = mad(brow02, (float4) a2.z, dot02);
            dot02 = mad(brow03, (float4) a2.w, dot02);
            dot12 = mad(brow10, (float4) a2.x, dot12);
            dot12 = mad(brow11, (float4) a2.y, dot12);
            dot12 = mad(brow12, (float4) a2.z, dot12);
            dot12 = mad(brow13, (float4) a2.w, dot12);

            const float4 a3 = slm[ i + 3 * TILE_K / VEC_SIZE ];
            dot03 = mad(brow00, (float4) a3.x, dot03);
            dot03 = mad(brow01, (float4) a3.y, dot03);
            dot03 = mad(brow02, (float4) a3.z, dot03);
            dot03 = mad(brow03, (float4) a3.w, dot03);
            dot13 = mad(brow10, (float4) a3.x, dot13);
            dot13 = mad(brow11, (float4) a3.y, dot13);
            dot13 = mad(brow12, (float4) a3.z, dot13);
            dot13 = mad(brow13, (float4) a3.w, dot13);

            const float4 a4 = slm[ i + 4 * TILE_K / VEC_SIZE ];
            dot04 = mad(brow00, (float4) a4.x, dot04);
            dot04 = mad(brow01, (float4) a4.y, dot04);
            dot04 = mad(brow02, (float4) a4.z, dot04);
            dot04 = mad(brow03, (float4) a4.w, dot04);
            dot14 = mad(brow10, (float4) a4.x, dot14);
            dot14 = mad(brow11, (float4) a4.y, dot14);
            dot14 = mad(brow12, (float4) a4.z, dot14);
            dot14 = mad(brow13, (float4) a4.w, dot14);

            const float4 a5 = slm[ i + 5 * TILE_K / VEC_SIZE ];
            dot05 = mad(brow00, (float4) a5.x, dot05);
            dot05 = mad(brow01, (float4) a5.y, dot05);
            dot05 = mad(brow02, (float4) a5.z, dot05);
            dot05 = mad(brow03, (float4) a5.w, dot05);
            dot15 = mad(brow10, (float4) a5.x, dot15);
            dot15 = mad(brow11, (float4) a5.y, dot15);
            dot15 = mad(brow12, (float4) a5.z, dot15);
            dot15 = mad(brow13, (float4) a5.w, dot15);

            const float4 a6 = slm[ i + 6 * TILE_K / VEC_SIZE ];
            dot06 = mad(brow00, (float4) a6.x, dot06);
            dot06 = mad(brow01, (float4) a6.y, dot06);
            dot06 = mad(brow02, (float4) a6.z, dot06);
            dot06 = mad(brow03, (float4) a6.w, dot06);
            dot16 = mad(brow10, (float4) a6.x, dot16);
            dot16 = mad(brow11, (float4) a6.y, dot16);
            dot16 = mad(brow12, (float4) a6.z, dot16);
            dot16 = mad(brow13, (float4) a6.w, dot16);

            const float4 a7 = slm[ i + 7 * TILE_K / VEC_SIZE ];
            dot07 = mad(brow00, (float4) a7.x, dot07);
            dot07 = mad(brow01, (float4) a7.y, dot07);
            dot07 = mad(brow02, (float4) a7.z, dot07);
            dot07 = mad(brow03, (float4) a7.w, dot07);
            dot17 = mad(brow10, (float4) a7.x, dot17);
            dot17 = mad(brow11, (float4) a7.y, dot17);
            dot17 = mad(brow12, (float4) a7.z, dot17);
            dot17 = mad(brow13, (float4) a7.w, dot17);

            i++;
        }
        while( i < TILE_K / VEC_SIZE );

        barrier(CLK_LOCAL_MEM_FENCE);

        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    __global float4 *dst_write1 = dst_write0 + ( TILE_N / 2 / VEC_SIZE );

    dst_write0[0] = dot00;  dst_write0 += width1;
    dst_write0[0] = dot01;  dst_write0 += width1;
    dst_write0[0] = dot02;  dst_write0 += width1;
    dst_write0[0] = dot03;  dst_write0 += width1;
    dst_write0[0] = dot04;  dst_write0 += width1;
    dst_write0[0] = dot05;  dst_write0 += width1;
    dst_write0[0] = dot06;  dst_write0 += width1;
    dst_write0[0] = dot07;  dst_write0 += width1;

    dst_write1[0] = dot10;  dst_write1 += width1;
    dst_write1[0] = dot11;  dst_write1 += width1;
    dst_write1[0] = dot12;  dst_write1 += width1;
    dst_write1[0] = dot13;  dst_write1 += width1;
    dst_write1[0] = dot14;  dst_write1 += width1;
    dst_write1[0] = dot15;  dst_write1 += width1;
    dst_write1[0] = dot16;  dst_write1 += width1;
    dst_write1[0] = dot17;  dst_write1 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N
#undef ROWS_PER_WI

////////////////////////////////////////////////////////////////
// L3_SLM_Fast_16x4_2

#define VEC_SIZE        4
#define TILE_M          32
#define TILE_K          64
#define TILE_N          128
#define ROWS_PER_WI     8

__attribute__((reqd_work_group_size(16, 4, 1)))
__kernel void L3_SLM_8x8_4x16(
    const __global float4 *src0,
    const __global float4 *src1,
    __global float4 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    // Src0 atile is M rows x K columns
    // M = 32
    // K = 64 = 16 float4s
    // requires sizeof(float) x M x K = 8K SLM
    __local float4 atile[TILE_M * TILE_K / VEC_SIZE];

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
    const int local_y = get_local_id(1);

    // Result ctile is M rows x N columns
    // M = 32, we have 4 rows of work-items, so we need 32/4 8 results down
    // N = 128, we have 16 columns of work-items, so we need 128/16 = 8 results across = 2 float4s across

    float4 dot00 = (float4)(0.f);
    float4 dot01 = (float4)(0.f);
    float4 dot02 = (float4)(0.f);
    float4 dot03 = (float4)(0.f);
    float4 dot04 = (float4)(0.f);
    float4 dot05 = (float4)(0.f);
    float4 dot06 = (float4)(0.f);
    float4 dot07 = (float4)(0.f);

    float4 dot10 = (float4)(0.f);
    float4 dot11 = (float4)(0.f);
    float4 dot12 = (float4)(0.f);
    float4 dot13 = (float4)(0.f);
    float4 dot14 = (float4)(0.f);
    float4 dot15 = (float4)(0.f);
    float4 dot16 = (float4)(0.f);
    float4 dot17 = (float4)(0.f);

    __global float4 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( ( group_y * TILE_M ) + ROWS_PER_WI * local_y ) * width1;

    // Src0 is used to load atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    const __global float4 *src0_read = src0 + local_x + ( ( group_y * TILE_M ) + ROWS_PER_WI * local_y ) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    // K = 64, we'll process four rows at a time
    // N = 128, we have 16 columns of work-items, so we need 128/16 = 8 floats across = 2 float4s across
    const __global float4 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );
    const __global float4 *src1_read1 = src1_read0 + ( TILE_N / 2 / VEC_SIZE );

    __local float4 *slm = atile + local_y * ( ROWS_PER_WI * TILE_K / VEC_SIZE );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        // We want to load atile, which is M rows x K columns
        // M = 32, and we have 4 rows of work-items, so each work-item must load 32/4 = 8 rows.
        // K = 64, and we have 16 columns of work-items, so each work-item must load 64/16 = 4 columns = 1 float4.
        slm[local_x + 0 * TILE_K / VEC_SIZE] = src0_read[0 * width0];
        slm[local_x + 1 * TILE_K / VEC_SIZE] = src0_read[1 * width0];
        slm[local_x + 2 * TILE_K / VEC_SIZE] = src0_read[2 * width0];
        slm[local_x + 3 * TILE_K / VEC_SIZE] = src0_read[3 * width0];
        slm[local_x + 4 * TILE_K / VEC_SIZE] = src0_read[4 * width0];
        slm[local_x + 5 * TILE_K / VEC_SIZE] = src0_read[5 * width0];
        slm[local_x + 6 * TILE_K / VEC_SIZE] = src0_read[6 * width0];
        slm[local_x + 7 * TILE_K / VEC_SIZE] = src0_read[7 * width0];

        src0_read += TILE_K / VEC_SIZE;

        barrier(CLK_LOCAL_MEM_FENCE);

        int i = 0;
        do
        {
            // We get better performance by loading btile first.
            const float4 brow00 = src1_read0[0];   src1_read0 += width1;
            const float4 brow01 = src1_read0[0];   src1_read0 += width1;
            const float4 brow02 = src1_read0[0];   src1_read0 += width1;
            const float4 brow03 = src1_read0[0];   src1_read0 += width1;
            const float4 brow10 = src1_read1[0];   src1_read1 += width1;
            const float4 brow11 = src1_read1[0];   src1_read1 += width1;
            const float4 brow12 = src1_read1[0];   src1_read1 += width1;
            const float4 brow13 = src1_read1[0];   src1_read1 += width1;

            const float4 a0 = slm[ i + 0 * TILE_K / VEC_SIZE ];
            dot00 = mad(brow00, (float4) a0.x, dot00);
            dot00 = mad(brow01, (float4) a0.y, dot00);
            dot00 = mad(brow02, (float4) a0.z, dot00);
            dot00 = mad(brow03, (float4) a0.w, dot00);
            dot10 = mad(brow10, (float4) a0.x, dot10);
            dot10 = mad(brow11, (float4) a0.y, dot10);
            dot10 = mad(brow12, (float4) a0.z, dot10);
            dot10 = mad(brow13, (float4) a0.w, dot10);

            const float4 a1 = slm[ i + 1 * TILE_K / VEC_SIZE ];
            dot01 = mad(brow00, (float4) a1.x, dot01);
            dot01 = mad(brow01, (float4) a1.y, dot01);
            dot01 = mad(brow02, (float4) a1.z, dot01);
            dot01 = mad(brow03, (float4) a1.w, dot01);
            dot11 = mad(brow10, (float4) a1.x, dot11);
            dot11 = mad(brow11, (float4) a1.y, dot11);
            dot11 = mad(brow12, (float4) a1.z, dot11);
            dot11 = mad(brow13, (float4) a1.w, dot11);

            const float4 a2 = slm[ i + 2 * TILE_K / VEC_SIZE ];
            dot02 = mad(brow00, (float4) a2.x, dot02);
            dot02 = mad(brow01, (float4) a2.y, dot02);
            dot02 = mad(brow02, (float4) a2.z, dot02);
            dot02 = mad(brow03, (float4) a2.w, dot02);
            dot12 = mad(brow10, (float4) a2.x, dot12);
            dot12 = mad(brow11, (float4) a2.y, dot12);
            dot12 = mad(brow12, (float4) a2.z, dot12);
            dot12 = mad(brow13, (float4) a2.w, dot12);

            const float4 a3 = slm[ i + 3 * TILE_K / VEC_SIZE ];
            dot03 = mad(brow00, (float4) a3.x, dot03);
            dot03 = mad(brow01, (float4) a3.y, dot03);
            dot03 = mad(brow02, (float4) a3.z, dot03);
            dot03 = mad(brow03, (float4) a3.w, dot03);
            dot13 = mad(brow10, (float4) a3.x, dot13);
            dot13 = mad(brow11, (float4) a3.y, dot13);
            dot13 = mad(brow12, (float4) a3.z, dot13);
            dot13 = mad(brow13, (float4) a3.w, dot13);

            const float4 a4 = slm[ i + 4 * TILE_K / VEC_SIZE ];
            dot04 = mad(brow00, (float4) a4.x, dot04);
            dot04 = mad(brow01, (float4) a4.y, dot04);
            dot04 = mad(brow02, (float4) a4.z, dot04);
            dot04 = mad(brow03, (float4) a4.w, dot04);
            dot14 = mad(brow10, (float4) a4.x, dot14);
            dot14 = mad(brow11, (float4) a4.y, dot14);
            dot14 = mad(brow12, (float4) a4.z, dot14);
            dot14 = mad(brow13, (float4) a4.w, dot14);

            const float4 a5 = slm[ i + 5 * TILE_K / VEC_SIZE ];
            dot05 = mad(brow00, (float4) a5.x, dot05);
            dot05 = mad(brow01, (float4) a5.y, dot05);
            dot05 = mad(brow02, (float4) a5.z, dot05);
            dot05 = mad(brow03, (float4) a5.w, dot05);
            dot15 = mad(brow10, (float4) a5.x, dot15);
            dot15 = mad(brow11, (float4) a5.y, dot15);
            dot15 = mad(brow12, (float4) a5.z, dot15);
            dot15 = mad(brow13, (float4) a5.w, dot15);

            const float4 a6 = slm[ i + 6 * TILE_K / VEC_SIZE ];
            dot06 = mad(brow00, (float4) a6.x, dot06);
            dot06 = mad(brow01, (float4) a6.y, dot06);
            dot06 = mad(brow02, (float4) a6.z, dot06);
            dot06 = mad(brow03, (float4) a6.w, dot06);
            dot16 = mad(brow10, (float4) a6.x, dot16);
            dot16 = mad(brow11, (float4) a6.y, dot16);
            dot16 = mad(brow12, (float4) a6.z, dot16);
            dot16 = mad(brow13, (float4) a6.w, dot16);

            const float4 a7 = slm[ i + 7 * TILE_K / VEC_SIZE ];
            dot07 = mad(brow00, (float4) a7.x, dot07);
            dot07 = mad(brow01, (float4) a7.y, dot07);
            dot07 = mad(brow02, (float4) a7.z, dot07);
            dot07 = mad(brow03, (float4) a7.w, dot07);
            dot17 = mad(brow10, (float4) a7.x, dot17);
            dot17 = mad(brow11, (float4) a7.y, dot17);
            dot17 = mad(brow12, (float4) a7.z, dot17);
            dot17 = mad(brow13, (float4) a7.w, dot17);

            i++;
        }
        while( i < TILE_K / VEC_SIZE );

        barrier(CLK_LOCAL_MEM_FENCE);

        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    __global float4 *dst_write1 = dst_write0 + ( TILE_N / 2 / VEC_SIZE );

    dst_write0[0] = dot00;  dst_write0 += width1;
    dst_write0[0] = dot01;  dst_write0 += width1;
    dst_write0[0] = dot02;  dst_write0 += width1;
    dst_write0[0] = dot03;  dst_write0 += width1;
    dst_write0[0] = dot04;  dst_write0 += width1;
    dst_write0[0] = dot05;  dst_write0 += width1;
    dst_write0[0] = dot06;  dst_write0 += width1;
    dst_write0[0] = dot07;  dst_write0 += width1;

    dst_write1[0] = dot10;  dst_write1 += width1;
    dst_write1[0] = dot11;  dst_write1 += width1;
    dst_write1[0] = dot12;  dst_write1 += width1;
    dst_write1[0] = dot13;  dst_write1 += width1;
    dst_write1[0] = dot14;  dst_write1 += width1;
    dst_write1[0] = dot15;  dst_write1 += width1;
    dst_write1[0] = dot16;  dst_write1 += width1;
    dst_write1[0] = dot17;  dst_write1 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N
#undef ROWS_PER_WI


////////////////////////////////////////////////////////////////
// L3_SLM_Fast_16x8_2

#define VEC_SIZE        4
#define TILE_M          64
#define TILE_K          64
#define TILE_N          128
#define ROWS_PER_WI     8

__attribute__((reqd_work_group_size(16, 8, 1)))
__kernel void L3_SLM_8x8_8x16(
    const __global float4 *src0,
    const __global float4 *src1,
    __global float4 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    // Src0 atile is M rows x K columns
    // M = 64
    // K = 64 = 16 float4s
    // requires sizeof(float) x M x K = 16K SLM
    __local float4 atile[TILE_M * TILE_K / VEC_SIZE];

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
    const int local_y = get_local_id(1);

    // Result ctile is M rows x N columns
    // M = 64, we have 8 rows of work-items, so we need 64/8 8 results down
    // N = 128, we have 16 columns of work-items, so we need 128/16 = 8 results across = 2 float4s across

    float4 dot00 = (float4)(0.f);
    float4 dot01 = (float4)(0.f);
    float4 dot02 = (float4)(0.f);
    float4 dot03 = (float4)(0.f);
    float4 dot04 = (float4)(0.f);
    float4 dot05 = (float4)(0.f);
    float4 dot06 = (float4)(0.f);
    float4 dot07 = (float4)(0.f);

    float4 dot10 = (float4)(0.f);
    float4 dot11 = (float4)(0.f);
    float4 dot12 = (float4)(0.f);
    float4 dot13 = (float4)(0.f);
    float4 dot14 = (float4)(0.f);
    float4 dot15 = (float4)(0.f);
    float4 dot16 = (float4)(0.f);
    float4 dot17 = (float4)(0.f);

    __global float4 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( ( group_y * TILE_M ) + ROWS_PER_WI * local_y ) * width1;

    // Src0 is used to load atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    const __global float4 *src0_read = src0 + local_x + ( ( group_y * TILE_M ) + ROWS_PER_WI * local_y ) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    // K = 64, we'll process four rows at a time
    // N = 128, we have 16 columns of work-items, so we need 128/16 = 8 floats across = 2 float4s across
    const __global float4 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );
    const __global float4 *src1_read1 = src1_read0 + ( TILE_N / 2 / VEC_SIZE );
  
    __local float4 *slm = atile + local_y * ( ROWS_PER_WI * TILE_K / VEC_SIZE );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        // We want to load atile, which is M rows x K columns
        // M = 64, and we have 8 rows of work-items, so each work-item must load 64/8 = 8 rows.
        // K = 64, and we have 16 columns of work-items, so each work-item must load 64/16 = 4 columns = 1 float4.
        slm[local_x + 0 * TILE_K / VEC_SIZE] = src0_read[0 * width0];
        slm[local_x + 1 * TILE_K / VEC_SIZE] = src0_read[1 * width0];
        slm[local_x + 2 * TILE_K / VEC_SIZE] = src0_read[2 * width0];
        slm[local_x + 3 * TILE_K / VEC_SIZE] = src0_read[3 * width0];
        slm[local_x + 4 * TILE_K / VEC_SIZE] = src0_read[4 * width0];
        slm[local_x + 5 * TILE_K / VEC_SIZE] = src0_read[5 * width0];
        slm[local_x + 6 * TILE_K / VEC_SIZE] = src0_read[6 * width0];
        slm[local_x + 7 * TILE_K / VEC_SIZE] = src0_read[7 * width0];
        src0_read += TILE_K / VEC_SIZE;

        barrier(CLK_LOCAL_MEM_FENCE);

        int i = 0;
        do
        {
            // We get better performance by loading btile first.
            const float4 brow00 = src1_read0[0];   src1_read0 += width1;
            const float4 brow01 = src1_read0[0];   src1_read0 += width1;
            const float4 brow02 = src1_read0[0];   src1_read0 += width1;
            const float4 brow03 = src1_read0[0];   src1_read0 += width1;
            const float4 brow10 = src1_read1[0];   src1_read1 += width1;
            const float4 brow11 = src1_read1[0];   src1_read1 += width1;
            const float4 brow12 = src1_read1[0];   src1_read1 += width1;
            const float4 brow13 = src1_read1[0];   src1_read1 += width1;

            const float4 a0 = slm[ i + 0 * TILE_K / VEC_SIZE ];
            dot00 = mad(brow00, (float4) a0.x, dot00);
            dot00 = mad(brow01, (float4) a0.y, dot00);
            dot00 = mad(brow02, (float4) a0.z, dot00);
            dot00 = mad(brow03, (float4) a0.w, dot00);
            dot10 = mad(brow10, (float4) a0.x, dot10);
            dot10 = mad(brow11, (float4) a0.y, dot10);
            dot10 = mad(brow12, (float4) a0.z, dot10);
            dot10 = mad(brow13, (float4) a0.w, dot10);

            const float4 a1 = slm[ i + 1 * TILE_K / VEC_SIZE ];
            dot01 = mad(brow00, (float4) a1.x, dot01);
            dot01 = mad(brow01, (float4) a1.y, dot01);
            dot01 = mad(brow02, (float4) a1.z, dot01);
            dot01 = mad(brow03, (float4) a1.w, dot01);
            dot11 = mad(brow10, (float4) a1.x, dot11);
            dot11 = mad(brow11, (float4) a1.y, dot11);
            dot11 = mad(brow12, (float4) a1.z, dot11);
            dot11 = mad(brow13, (float4) a1.w, dot11);

            const float4 a2 = slm[ i + 2 * TILE_K / VEC_SIZE ];
            dot02 = mad(brow00, (float4) a2.x, dot02);
            dot02 = mad(brow01, (float4) a2.y, dot02);
            dot02 = mad(brow02, (float4) a2.z, dot02);
            dot02 = mad(brow03, (float4) a2.w, dot02);
            dot12 = mad(brow10, (float4) a2.x, dot12);
            dot12 = mad(brow11, (float4) a2.y, dot12);
            dot12 = mad(brow12, (float4) a2.z, dot12);
            dot12 = mad(brow13, (float4) a2.w, dot12);

            const float4 a3 = slm[ i + 3 * TILE_K / VEC_SIZE ];
            dot03 = mad(brow00, (float4) a3.x, dot03);
            dot03 = mad(brow01, (float4) a3.y, dot03);
            dot03 = mad(brow02, (float4) a3.z, dot03);
            dot03 = mad(brow03, (float4) a3.w, dot03);
            dot13 = mad(brow10, (float4) a3.x, dot13);
            dot13 = mad(brow11, (float4) a3.y, dot13);
            dot13 = mad(brow12, (float4) a3.z, dot13);
            dot13 = mad(brow13, (float4) a3.w, dot13);

            const float4 a4 = slm[ i + 4 * TILE_K / VEC_SIZE ];
            dot04 = mad(brow00, (float4) a4.x, dot04);
            dot04 = mad(brow01, (float4) a4.y, dot04);
            dot04 = mad(brow02, (float4) a4.z, dot04);
            dot04 = mad(brow03, (float4) a4.w, dot04);
            dot14 = mad(brow10, (float4) a4.x, dot14);
            dot14 = mad(brow11, (float4) a4.y, dot14);
            dot14 = mad(brow12, (float4) a4.z, dot14);
            dot14 = mad(brow13, (float4) a4.w, dot14);

            const float4 a5 = slm[ i + 5 * TILE_K / VEC_SIZE ];
            dot05 = mad(brow00, (float4) a5.x, dot05);
            dot05 = mad(brow01, (float4) a5.y, dot05);
            dot05 = mad(brow02, (float4) a5.z, dot05);
            dot05 = mad(brow03, (float4) a5.w, dot05);
            dot15 = mad(brow10, (float4) a5.x, dot15);
            dot15 = mad(brow11, (float4) a5.y, dot15);
            dot15 = mad(brow12, (float4) a5.z, dot15);
            dot15 = mad(brow13, (float4) a5.w, dot15);

            const float4 a6 = slm[ i + 6 * TILE_K / VEC_SIZE ];
            dot06 = mad(brow00, (float4) a6.x, dot06);
            dot06 = mad(brow01, (float4) a6.y, dot06);
            dot06 = mad(brow02, (float4) a6.z, dot06);
            dot06 = mad(brow03, (float4) a6.w, dot06);
            dot16 = mad(brow10, (float4) a6.x, dot16);
            dot16 = mad(brow11, (float4) a6.y, dot16);
            dot16 = mad(brow12, (float4) a6.z, dot16);
            dot16 = mad(brow13, (float4) a6.w, dot16);

            const float4 a7 = slm[ i + 7 * TILE_K / VEC_SIZE ];
            dot07 = mad(brow00, (float4) a7.x, dot07);
            dot07 = mad(brow01, (float4) a7.y, dot07);
            dot07 = mad(brow02, (float4) a7.z, dot07);
            dot07 = mad(brow03, (float4) a7.w, dot07);
            dot17 = mad(brow10, (float4) a7.x, dot17);
            dot17 = mad(brow11, (float4) a7.y, dot17);
            dot17 = mad(brow12, (float4) a7.z, dot17);
            dot17 = mad(brow13, (float4) a7.w, dot17);

            i++;
        }
        while( i < TILE_K / VEC_SIZE );

        barrier(CLK_LOCAL_MEM_FENCE);

        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    __global float4 *dst_write1 = dst_write0 + ( TILE_N / 2 / VEC_SIZE );

    dst_write0[0] = dot00;  dst_write0 += width1;
    dst_write0[0] = dot01;  dst_write0 += width1;
    dst_write0[0] = dot02;  dst_write0 += width1;
    dst_write0[0] = dot03;  dst_write0 += width1;
    dst_write0[0] = dot04;  dst_write0 += width1;
    dst_write0[0] = dot05;  dst_write0 += width1;
    dst_write0[0] = dot06;  dst_write0 += width1;
    dst_write0[0] = dot07;  dst_write0 += width1;

    dst_write1[0] = dot10;  dst_write1 += width1;
    dst_write1[0] = dot11;  dst_write1 += width1;
    dst_write1[0] = dot12;  dst_write1 += width1;
    dst_write1[0] = dot13;  dst_write1 += width1;
    dst_write1[0] = dot14;  dst_write1 += width1;
    dst_write1[0] = dot15;  dst_write1 += width1;
    dst_write1[0] = dot16;  dst_write1 += width1;
    dst_write1[0] = dot17;  dst_write1 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N
#undef ROWS_PER_WI

#ifdef GPU
////////////////////////////////////////////////////////////////
// L3_SIMD_1x4

#define TILE_M          4
#define TILE_K          8
#define TILE_N          8

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void L3_SIMD_1x4(
    const __global float *src0,
    const __global float *src1,
    __global float *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 4, we have 1 row of work-items, so we need 4/1 = 4 results down
    // N = 8, we have 8 columns of work-items, so we need 8/8 = 1 result across

    float   dot00 = 0.f;
    float   dot01 = 0.f;
    float   dot02 = 0.f;
    float   dot03 = 0.f;

    __global float *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M ) * width1;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    const __global float *src0_read = src0 + local_x + ( group_y * TILE_M ) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    const __global float *src1_read0 = src1 + local_x + ( group_x * ( TILE_N ) );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        // We want to load atile, which is M rows x K columns
        // M = 4, we have 1 row of work-items, so each work-item must load 4/1 = 4 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        float   arow;

        // Now load btile, which is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        float   brow0 = src1_read0[0];  src1_read0 += width1;
        float   brow1 = src1_read0[0];  src1_read0 += width1;
        float   brow2 = src1_read0[0];  src1_read0 += width1;
        float   brow3 = src1_read0[0];  src1_read0 += width1;
        float   brow4 = src1_read0[0];  src1_read0 += width1;
        float   brow5 = src1_read0[0];  src1_read0 += width1;
        float   brow6 = src1_read0[0];  src1_read0 += width1;
        float   brow7 = src1_read0[0];  src1_read0 += width1;

        arow = src0_read[ 0 * width0 ];
        dot00 = mad(intel_sub_group_shuffle( arow, 0 ), brow0, dot00);
        dot00 = mad(intel_sub_group_shuffle( arow, 1 ), brow1, dot00);
        dot00 = mad(intel_sub_group_shuffle( arow, 2 ), brow2, dot00);
        dot00 = mad(intel_sub_group_shuffle( arow, 3 ), brow3, dot00);
        dot00 = mad(intel_sub_group_shuffle( arow, 4 ), brow4, dot00);
        dot00 = mad(intel_sub_group_shuffle( arow, 5 ), brow5, dot00);
        dot00 = mad(intel_sub_group_shuffle( arow, 6 ), brow6, dot00);
        dot00 = mad(intel_sub_group_shuffle( arow, 7 ), brow7, dot00);

        arow = src0_read[ 1 * width0 ];
        dot01 = mad(intel_sub_group_shuffle( arow, 0 ), brow0, dot01);
        dot01 = mad(intel_sub_group_shuffle( arow, 1 ), brow1, dot01);
        dot01 = mad(intel_sub_group_shuffle( arow, 2 ), brow2, dot01);
        dot01 = mad(intel_sub_group_shuffle( arow, 3 ), brow3, dot01);
        dot01 = mad(intel_sub_group_shuffle( arow, 5 ), brow5, dot01);
        dot01 = mad(intel_sub_group_shuffle( arow, 4 ), brow4, dot01);
        dot01 = mad(intel_sub_group_shuffle( arow, 6 ), brow6, dot01);
        dot01 = mad(intel_sub_group_shuffle( arow, 7 ), brow7, dot01);

        arow = src0_read[ 2 * width0 ];
        dot02 = mad(intel_sub_group_shuffle( arow, 0 ), brow0, dot02);
        dot02 = mad(intel_sub_group_shuffle( arow, 1 ), brow1, dot02);
        dot02 = mad(intel_sub_group_shuffle( arow, 2 ), brow2, dot02);
        dot02 = mad(intel_sub_group_shuffle( arow, 3 ), brow3, dot02);
        dot02 = mad(intel_sub_group_shuffle( arow, 4 ), brow4, dot02);
        dot02 = mad(intel_sub_group_shuffle( arow, 5 ), brow5, dot02);
        dot02 = mad(intel_sub_group_shuffle( arow, 6 ), brow6, dot02);
        dot02 = mad(intel_sub_group_shuffle( arow, 7 ), brow7, dot02);

        arow = src0_read[ 3 * width0 ];
        dot03 = mad(intel_sub_group_shuffle( arow, 0 ), brow0, dot03);
        dot03 = mad(intel_sub_group_shuffle( arow, 1 ), brow1, dot03);
        dot03 = mad(intel_sub_group_shuffle( arow, 2 ), brow2, dot03);
        dot03 = mad(intel_sub_group_shuffle( arow, 3 ), brow3, dot03);
        dot03 = mad(intel_sub_group_shuffle( arow, 4 ), brow4, dot03);
        dot03 = mad(intel_sub_group_shuffle( arow, 5 ), brow5, dot03);
        dot03 = mad(intel_sub_group_shuffle( arow, 6 ), brow6, dot03);
        dot03 = mad(intel_sub_group_shuffle( arow, 7 ), brow7, dot03);

        src0_read += TILE_K;
        w += TILE_K;
    }
    while( w < width0 );

    dst_write0[0] = dot00;  dst_write0 += width1;
    dst_write0[0] = dot01;  dst_write0 += width1;
    dst_write0[0] = dot02;  dst_write0 += width1;
    dst_write0[0] = dot03;  dst_write0 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N


#define VEC_SIZE        2
#define TILE_M          64
#define TILE_K          16
#define TILE_N          16
__attribute__((reqd_work_group_size(8, 4, 1)))
__kernel void L3_SIMD_16x2_4x8(
    const __global float2 *src0,
    const __global float2 *src1,
    __global float2 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);

    // Result ctile is M rows x N columns
    // M = 16, we have 1 row of work-items, so we need 16/1 = 16 results down
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 result across

    float2  dot00 = 0.f;
    float2  dot01 = 0.f;
    float2  dot02 = 0.f;
    float2  dot03 = 0.f;
    float2  dot04 = 0.f;
    float2  dot05 = 0.f;
    float2  dot06 = 0.f;
    float2  dot07 = 0.f;
    float2  dot08 = 0.f;
    float2  dot09 = 0.f;
    float2  dot0a = 0.f;
    float2  dot0b = 0.f;
    float2  dot0c = 0.f;
    float2  dot0d = 0.f;
    float2  dot0e = 0.f;
    float2  dot0f = 0.f;

    __global float2 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( group_y * TILE_M + local_y * 16) * width1;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    const __global float2 *src0_read = src0 + local_x + ( group_y * TILE_M + local_y * 16) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    const __global float2 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        // We want to load atile, which is M rows x K columns
        // M = 16, we have 1 row of work-items, so each work-item must load 16/1 = 16 rows
        // K = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 columns
        float2  arow;

        // Now load btile, which is K rows x N columns
        // K = 16, we have 1 row of work-items, so each work-item must load 16/1 = 16 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 columns
        float2  brow0 = src1_read0[0];  src1_read0 += width1;
        float2  brow1 = src1_read0[0];  src1_read0 += width1;
        float2  brow2 = src1_read0[0];  src1_read0 += width1;
        float2  brow3 = src1_read0[0];  src1_read0 += width1;
        float2  brow4 = src1_read0[0];  src1_read0 += width1;
        float2  brow5 = src1_read0[0];  src1_read0 += width1;
        float2  brow6 = src1_read0[0];  src1_read0 += width1;
        float2  brow7 = src1_read0[0];  src1_read0 += width1;
        float2  brow8 = src1_read0[0];  src1_read0 += width1;
        float2  brow9 = src1_read0[0];  src1_read0 += width1;
        float2  browa = src1_read0[0];  src1_read0 += width1;
        float2  browb = src1_read0[0];  src1_read0 += width1;
        float2  browc = src1_read0[0];  src1_read0 += width1;
        float2  browd = src1_read0[0];  src1_read0 += width1;
        float2  browe = src1_read0[0];  src1_read0 += width1;
        float2  browf = src1_read0[0];  src1_read0 += width1;

#define MM_DOT_PRODUCT( _row, _dot )   \
        arow = src0_read[ _row * width0 ];                           \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 0 ).x), brow0, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 0 ).y), brow1, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 1 ).x), brow2, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 1 ).y), brow3, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 2 ).x), brow4, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 2 ).y), brow5, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 3 ).x), brow6, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 3 ).y), brow7, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 4 ).x), brow8, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 4 ).y), brow9, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 5 ).x), browa, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 5 ).y), browb, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 6 ).x), browc, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 6 ).y), browd, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 7 ).x), browe, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 7 ).y), browf, _dot );

        MM_DOT_PRODUCT( 0x0, dot00 );
        MM_DOT_PRODUCT( 0x1, dot01 );
        MM_DOT_PRODUCT( 0x2, dot02 );
        MM_DOT_PRODUCT( 0x3, dot03 );
        MM_DOT_PRODUCT( 0x4, dot04 );
        MM_DOT_PRODUCT( 0x5, dot05 );
        MM_DOT_PRODUCT( 0x6, dot06 );
        MM_DOT_PRODUCT( 0x7, dot07 );
        MM_DOT_PRODUCT( 0x8, dot08 );
        MM_DOT_PRODUCT( 0x9, dot09 );
        MM_DOT_PRODUCT( 0xa, dot0a );
        MM_DOT_PRODUCT( 0xb, dot0b );
        MM_DOT_PRODUCT( 0xc, dot0c );
        MM_DOT_PRODUCT( 0xd, dot0d );
        MM_DOT_PRODUCT( 0xe, dot0e );
        MM_DOT_PRODUCT( 0xf, dot0f );

#undef MM_DOT_PRODUCT

        src0_read += TILE_K / VEC_SIZE;
        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    dst_write0[0] = dot00;  dst_write0 += width1;
    dst_write0[0] = dot01;  dst_write0 += width1;
    dst_write0[0] = dot02;  dst_write0 += width1;
    dst_write0[0] = dot03;  dst_write0 += width1;
    dst_write0[0] = dot04;  dst_write0 += width1;
    dst_write0[0] = dot05;  dst_write0 += width1;
    dst_write0[0] = dot06;  dst_write0 += width1;
    dst_write0[0] = dot07;  dst_write0 += width1;
    dst_write0[0] = dot08;  dst_write0 += width1;
    dst_write0[0] = dot09;  dst_write0 += width1;
    dst_write0[0] = dot0a;  dst_write0 += width1;
    dst_write0[0] = dot0b;  dst_write0 += width1;
    dst_write0[0] = dot0c;  dst_write0 += width1;
    dst_write0[0] = dot0d;  dst_write0 += width1;
    dst_write0[0] = dot0e;  dst_write0 += width1;
    dst_write0[0] = dot0f;  dst_write0 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N

////////////////////////////////////////////////////////////////
// L3_SIMD_2x16
#define VEC_SIZE        2
#define TILE_M          16
#define TILE_K          16
#define TILE_N          16

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void L3_SIMD_16x2_1x8(
    const __global float2 *src0,
    const __global float2 *src1,
    __global float2 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 16, we have 1 row of work-items, so we need 16/1 = 16 results down
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 result across

    float2  dot00 = 0.f;
    float2  dot01 = 0.f;
    float2  dot02 = 0.f;
    float2  dot03 = 0.f;
    float2  dot04 = 0.f;
    float2  dot05 = 0.f;
    float2  dot06 = 0.f;
    float2  dot07 = 0.f;
    float2  dot08 = 0.f;
    float2  dot09 = 0.f;
    float2  dot0a = 0.f;
    float2  dot0b = 0.f;
    float2  dot0c = 0.f;
    float2  dot0d = 0.f;
    float2  dot0e = 0.f;
    float2  dot0f = 0.f;

    __global float2 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( group_y * TILE_M ) * width1;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    const __global float2 *src0_read = src0 + local_x + ( group_y * TILE_M ) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    const __global float2 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        // We want to load atile, which is M rows x K columns
        // M = 16, we have 1 row of work-items, so each work-item must load 16/1 = 16 rows
        // K = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 columns
        float2  arow;

        // Now load btile, which is K rows x N columns
        // K = 16, we have 1 row of work-items, so each work-item must load 16/1 = 16 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 columns
        float2  brow0 = src1_read0[0];  src1_read0 += width1;
        float2  brow1 = src1_read0[0];  src1_read0 += width1;
        float2  brow2 = src1_read0[0];  src1_read0 += width1;
        float2  brow3 = src1_read0[0];  src1_read0 += width1;
        float2  brow4 = src1_read0[0];  src1_read0 += width1;
        float2  brow5 = src1_read0[0];  src1_read0 += width1;
        float2  brow6 = src1_read0[0];  src1_read0 += width1;
        float2  brow7 = src1_read0[0];  src1_read0 += width1;
        float2  brow8 = src1_read0[0];  src1_read0 += width1;
        float2  brow9 = src1_read0[0];  src1_read0 += width1;
        float2  browa = src1_read0[0];  src1_read0 += width1;
        float2  browb = src1_read0[0];  src1_read0 += width1;
        float2  browc = src1_read0[0];  src1_read0 += width1;
        float2  browd = src1_read0[0];  src1_read0 += width1;
        float2  browe = src1_read0[0];  src1_read0 += width1;
        float2  browf = src1_read0[0];  src1_read0 += width1;

#define MM_DOT_PRODUCT( _row, _dot )   \
        arow = src0_read[ _row * width0 ];                           \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 0 ).x), brow0, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 0 ).y), brow1, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 1 ).x), brow2, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 1 ).y), brow3, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 2 ).x), brow4, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 2 ).y), brow5, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 3 ).x), brow6, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 3 ).y), brow7, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 4 ).x), brow8, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 4 ).y), brow9, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 5 ).x), browa, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 5 ).y), browb, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 6 ).x), browc, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 6 ).y), browd, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 7 ).x), browe, _dot ); \
        _dot = mad( (float2)(intel_sub_group_shuffle( arow, 7 ).y), browf, _dot );

        MM_DOT_PRODUCT( 0x0, dot00 );
        MM_DOT_PRODUCT( 0x1, dot01 );
        MM_DOT_PRODUCT( 0x2, dot02 );
        MM_DOT_PRODUCT( 0x3, dot03 );
        MM_DOT_PRODUCT( 0x4, dot04 );
        MM_DOT_PRODUCT( 0x5, dot05 );
        MM_DOT_PRODUCT( 0x6, dot06 );
        MM_DOT_PRODUCT( 0x7, dot07 );
        MM_DOT_PRODUCT( 0x8, dot08 );
        MM_DOT_PRODUCT( 0x9, dot09 );
        MM_DOT_PRODUCT( 0xa, dot0a );
        MM_DOT_PRODUCT( 0xb, dot0b );
        MM_DOT_PRODUCT( 0xc, dot0c );
        MM_DOT_PRODUCT( 0xd, dot0d );
        MM_DOT_PRODUCT( 0xe, dot0e );
        MM_DOT_PRODUCT( 0xf, dot0f );

#undef MM_DOT_PRODUCT

        src0_read += TILE_K / VEC_SIZE;
        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    dst_write0[0] = dot00;  dst_write0 += width1;
    dst_write0[0] = dot01;  dst_write0 += width1;
    dst_write0[0] = dot02;  dst_write0 += width1;
    dst_write0[0] = dot03;  dst_write0 += width1;
    dst_write0[0] = dot04;  dst_write0 += width1;
    dst_write0[0] = dot05;  dst_write0 += width1;
    dst_write0[0] = dot06;  dst_write0 += width1;
    dst_write0[0] = dot07;  dst_write0 += width1;
    dst_write0[0] = dot08;  dst_write0 += width1;
    dst_write0[0] = dot09;  dst_write0 += width1;
    dst_write0[0] = dot0a;  dst_write0 += width1;
    dst_write0[0] = dot0b;  dst_write0 += width1;
    dst_write0[0] = dot0c;  dst_write0 += width1;
    dst_write0[0] = dot0d;  dst_write0 += width1;
    dst_write0[0] = dot0e;  dst_write0 += width1;
    dst_write0[0] = dot0f;  dst_write0 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N


////////////////////////////////////////////////////////////////
// L3_SIMD_2x32

#define VEC_SIZE        2
#define TILE_M          32
#define TILE_K          16
#define TILE_N          16

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void L3_SIMD_32x2_1x8(
    const __global float2 *src0,
    const __global float2 *src1,
    __global float2 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 32, we have 1 row of work-items, so we need 32/1 = 32 results down
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 result across

    float2  dot00 = 0.f;
    float2  dot01 = 0.f;
    float2  dot02 = 0.f;
    float2  dot03 = 0.f;
    float2  dot04 = 0.f;
    float2  dot05 = 0.f;
    float2  dot06 = 0.f;
    float2  dot07 = 0.f;
    float2  dot08 = 0.f;
    float2  dot09 = 0.f;
    float2  dot0a = 0.f;
    float2  dot0b = 0.f;
    float2  dot0c = 0.f;
    float2  dot0d = 0.f;
    float2  dot0e = 0.f;
    float2  dot0f = 0.f;
    float2  dot10 = 0.f;
    float2  dot11 = 0.f;
    float2  dot12 = 0.f;
    float2  dot13 = 0.f;
    float2  dot14 = 0.f;
    float2  dot15 = 0.f;
    float2  dot16 = 0.f;
    float2  dot17 = 0.f;
    float2  dot18 = 0.f;
    float2  dot19 = 0.f;
    float2  dot1a = 0.f;
    float2  dot1b = 0.f;
    float2  dot1c = 0.f;
    float2  dot1d = 0.f;
    float2  dot1e = 0.f;
    float2  dot1f = 0.f;

    __global float2 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( group_y * TILE_M ) * width1;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    const __global float2 *src0_read = src0 + local_x + ( group_y * TILE_M ) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    const __global float2 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        // We want to load atile, which is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 columns
        float2  arow;
        
        // Now load btile, which is K rows x N columns
        // K = 16, we have 1 row of work-items, so each work-item must load 16/1 = 16 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 columns
        float2  brow0 = src1_read0[0];  src1_read0 += width1;
        float2  brow1 = src1_read0[0];  src1_read0 += width1;
        float2  brow2 = src1_read0[0];  src1_read0 += width1;
        float2  brow3 = src1_read0[0];  src1_read0 += width1;
        float2  brow4 = src1_read0[0];  src1_read0 += width1;
        float2  brow5 = src1_read0[0];  src1_read0 += width1;
        float2  brow6 = src1_read0[0];  src1_read0 += width1;
        float2  brow7 = src1_read0[0];  src1_read0 += width1;
        float2  brow8 = src1_read0[0];  src1_read0 += width1;
        float2  brow9 = src1_read0[0];  src1_read0 += width1;
        float2  browa = src1_read0[0];  src1_read0 += width1;
        float2  browb = src1_read0[0];  src1_read0 += width1;
        float2  browc = src1_read0[0];  src1_read0 += width1;
        float2  browd = src1_read0[0];  src1_read0 += width1;
        float2  browe = src1_read0[0];  src1_read0 += width1;
        float2  browf = src1_read0[0];  src1_read0 += width1;

#define MM_DOT_PRODUCT( _row, _dot ) \
        arow = src0_read[ _row * width0 ];                          \
        _dot = mad( intel_sub_group_shuffle( arow, 0 ).x, brow0, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 0 ).y, brow1, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 1 ).x, brow2, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 1 ).y, brow3, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 2 ).x, brow4, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 2 ).y, brow5, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 3 ).x, brow6, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 3 ).y, brow7, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 4 ).x, brow8, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 4 ).y, brow9, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 5 ).x, browa, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 5 ).y, browb, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 6 ).x, browc, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 6 ).y, browd, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 7 ).x, browe, _dot ); \
        _dot = mad( intel_sub_group_shuffle( arow, 7 ).y, browf, _dot );

        MM_DOT_PRODUCT( 0x00, dot00 );
        MM_DOT_PRODUCT( 0x01, dot01 );
        MM_DOT_PRODUCT( 0x02, dot02 );
        MM_DOT_PRODUCT( 0x03, dot03 );
        MM_DOT_PRODUCT( 0x04, dot04 );
        MM_DOT_PRODUCT( 0x05, dot05 );
        MM_DOT_PRODUCT( 0x06, dot06 );
        MM_DOT_PRODUCT( 0x07, dot07 );
        MM_DOT_PRODUCT( 0x08, dot08 );
        MM_DOT_PRODUCT( 0x09, dot09 );
        MM_DOT_PRODUCT( 0x0a, dot0a );
        MM_DOT_PRODUCT( 0x0b, dot0b );
        MM_DOT_PRODUCT( 0x0c, dot0c );
        MM_DOT_PRODUCT( 0x0d, dot0d );
        MM_DOT_PRODUCT( 0x0e, dot0e );
        MM_DOT_PRODUCT( 0x0f, dot0f );
        MM_DOT_PRODUCT( 0x10, dot10 );
        MM_DOT_PRODUCT( 0x11, dot11 );
        MM_DOT_PRODUCT( 0x12, dot12 );
        MM_DOT_PRODUCT( 0x13, dot13 );
        MM_DOT_PRODUCT( 0x14, dot14 );
        MM_DOT_PRODUCT( 0x15, dot15 );
        MM_DOT_PRODUCT( 0x16, dot16 );
        MM_DOT_PRODUCT( 0x17, dot17 );
        MM_DOT_PRODUCT( 0x18, dot18 );
        MM_DOT_PRODUCT( 0x19, dot19 );
        MM_DOT_PRODUCT( 0x1a, dot1a );
        MM_DOT_PRODUCT( 0x1b, dot1b );
        MM_DOT_PRODUCT( 0x1c, dot1c );
        MM_DOT_PRODUCT( 0x1d, dot1d );
        MM_DOT_PRODUCT( 0x1e, dot1e );
        MM_DOT_PRODUCT( 0x1f, dot1f );

#undef MM_DOT_PRODUCT

        src0_read += TILE_K / VEC_SIZE;
        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    dst_write0[0] = dot00;  dst_write0 += width1;
    dst_write0[0] = dot01;  dst_write0 += width1;
    dst_write0[0] = dot02;  dst_write0 += width1;
    dst_write0[0] = dot03;  dst_write0 += width1;
    dst_write0[0] = dot04;  dst_write0 += width1;
    dst_write0[0] = dot05;  dst_write0 += width1;
    dst_write0[0] = dot06;  dst_write0 += width1;
    dst_write0[0] = dot07;  dst_write0 += width1;
    dst_write0[0] = dot08;  dst_write0 += width1;
    dst_write0[0] = dot09;  dst_write0 += width1;
    dst_write0[0] = dot0a;  dst_write0 += width1;
    dst_write0[0] = dot0b;  dst_write0 += width1;
    dst_write0[0] = dot0c;  dst_write0 += width1;
    dst_write0[0] = dot0d;  dst_write0 += width1;
    dst_write0[0] = dot0e;  dst_write0 += width1;
    dst_write0[0] = dot0f;  dst_write0 += width1;
    dst_write0[0] = dot10;  dst_write0 += width1;
    dst_write0[0] = dot11;  dst_write0 += width1;
    dst_write0[0] = dot12;  dst_write0 += width1;
    dst_write0[0] = dot13;  dst_write0 += width1;
    dst_write0[0] = dot14;  dst_write0 += width1;
    dst_write0[0] = dot15;  dst_write0 += width1;
    dst_write0[0] = dot16;  dst_write0 += width1;
    dst_write0[0] = dot17;  dst_write0 += width1;
    dst_write0[0] = dot18;  dst_write0 += width1;
    dst_write0[0] = dot19;  dst_write0 += width1;
    dst_write0[0] = dot1a;  dst_write0 += width1;
    dst_write0[0] = dot1b;  dst_write0 += width1;
    dst_write0[0] = dot1c;  dst_write0 += width1;
    dst_write0[0] = dot1d;  dst_write0 += width1;
    dst_write0[0] = dot1e;  dst_write0 += width1;
    dst_write0[0] = dot1f;  dst_write0 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N

////////////////////////////////////////////////////////////////
// L3_SIMD_4x8

#define VEC_SIZE        4
#define TILE_M          8
#define TILE_K          32
#define TILE_N          32

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void L3_SIMD_8x4_1x8(
    const __global float4 *src0,
    const __global float4 *src1,
    __global float4 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 8, we have 1 rows of work-items, so we need 8/1 = 8 results down
    // N = 32, we have 8 columns of work-items, so we need 32/8 = 4 results across = 1 float4s across

    float4 dot00 = (float4)(0.f);
    float4 dot01 = (float4)(0.f);
    float4 dot02 = (float4)(0.f);
    float4 dot03 = (float4)(0.f);
    float4 dot04 = (float4)(0.f);
    float4 dot05 = (float4)(0.f);
    float4 dot06 = (float4)(0.f);
    float4 dot07 = (float4)(0.f);

    __global float4 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( group_y * TILE_M ) * width1;

    // Src0 is used directly as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    // M = 8, we have 1 rows of work-items, so we need 8/1 = 8 rows.
    // K = 32, we have 8 columns of work-items, so we need 32/8 = 4 floats across = 1 float4s across
    const __global float4 *src0_read = src0 + local_x + ( group_y * TILE_M ) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    // K = 32, we'll process four rows at a time
    // N = 32, we have 8 columns of work-items, so we need 32/8 = 4 floats across = 1 float4s across
    const __global float4 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        const float4 arow0 = src0_read[ 0 * width0 ];
        const float4 arow1 = src0_read[ 1 * width0 ];
        const float4 arow2 = src0_read[ 2 * width0 ];
        const float4 arow3 = src0_read[ 3 * width0 ];
        const float4 arow4 = src0_read[ 4 * width0 ];
        const float4 arow5 = src0_read[ 5 * width0 ];
        const float4 arow6 = src0_read[ 6 * width0 ];
        const float4 arow7 = src0_read[ 7 * width0 ];

#define ITERATION( _index ) \
        {   \
            const float4 a0 = intel_sub_group_shuffle( arow0, _index ); \
            const float4 a1 = intel_sub_group_shuffle( arow1, _index ); \
            const float4 a2 = intel_sub_group_shuffle( arow2, _index ); \
            const float4 a3 = intel_sub_group_shuffle( arow3, _index ); \
            const float4 a4 = intel_sub_group_shuffle( arow4, _index ); \
            const float4 a5 = intel_sub_group_shuffle( arow5, _index ); \
            const float4 a6 = intel_sub_group_shuffle( arow6, _index ); \
            const float4 a7 = intel_sub_group_shuffle( arow7, _index ); \
            const float4 brow00 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow01 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow02 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow03 = src1_read0[ 0 ];   src1_read0 += width1;    \
            dot00 = mad(brow00, (float4) a0.x, dot00);  \
            dot00 = mad(brow01, (float4) a0.y, dot00);  \
            dot00 = mad(brow02, (float4) a0.z, dot00);  \
            dot00 = mad(brow03, (float4) a0.w, dot00);  \
            dot01 = mad(brow00, (float4) a1.x, dot01);  \
            dot01 = mad(brow01, (float4) a1.y, dot01);  \
            dot01 = mad(brow02, (float4) a1.z, dot01);  \
            dot01 = mad(brow03, (float4) a1.w, dot01);  \
            dot02 = mad(brow00, (float4) a2.x, dot02);  \
            dot02 = mad(brow01, (float4) a2.y, dot02);  \
            dot02 = mad(brow02, (float4) a2.z, dot02);  \
            dot02 = mad(brow03, (float4) a2.w, dot02);  \
            dot03 = mad(brow00, (float4) a3.x, dot03);  \
            dot03 = mad(brow01, (float4) a3.y, dot03);  \
            dot03 = mad(brow02, (float4) a3.z, dot03);  \
            dot03 = mad(brow03, (float4) a3.w, dot03);  \
            dot04 = mad(brow00, (float4) a4.x, dot04);  \
            dot04 = mad(brow01, (float4) a4.y, dot04);  \
            dot04 = mad(brow02, (float4) a4.z, dot04);  \
            dot04 = mad(brow03, (float4) a4.w, dot04);  \
            dot05 = mad(brow00, (float4) a5.x, dot05);  \
            dot05 = mad(brow01, (float4) a5.y, dot05);  \
            dot05 = mad(brow02, (float4) a5.z, dot05);  \
            dot05 = mad(brow03, (float4) a5.w, dot05);  \
            dot06 = mad(brow00, (float4) a6.x, dot06);  \
            dot06 = mad(brow01, (float4) a6.y, dot06);  \
            dot06 = mad(brow02, (float4) a6.z, dot06);  \
            dot06 = mad(brow03, (float4) a6.w, dot06);  \
            dot07 = mad(brow00, (float4) a7.x, dot07);  \
            dot07 = mad(brow01, (float4) a7.y, dot07);  \
            dot07 = mad(brow02, (float4) a7.z, dot07);  \
            dot07 = mad(brow03, (float4) a7.w, dot07);  \
        }

        // We need K/VEC_SIZE iterations.
        // K = 32, VEC_SIZE = 4
        // So, 32/4 = 8 iterations.
        ITERATION( 0 );
        ITERATION( 1 );
        ITERATION( 2 );
        ITERATION( 3 );
        ITERATION( 4 );
        ITERATION( 5 );
        ITERATION( 6 );
        ITERATION( 7 );

#undef ITERATION

        src0_read += TILE_K / VEC_SIZE;
        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    dst_write0[ 0 ] = dot00;  dst_write0 += width1;
    dst_write0[ 0 ] = dot01;  dst_write0 += width1;
    dst_write0[ 0 ] = dot02;  dst_write0 += width1;
    dst_write0[ 0 ] = dot03;  dst_write0 += width1;
    dst_write0[ 0 ] = dot04;  dst_write0 += width1;
    dst_write0[ 0 ] = dot05;  dst_write0 += width1;
    dst_write0[ 0 ] = dot06;  dst_write0 += width1;
    dst_write0[ 0 ] = dot07;  dst_write0 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N


#define VEC_SIZE        4
#define TILE_M          64
#define TILE_K          32
#define TILE_N          32

__attribute__((reqd_work_group_size(8, 8, 1)))
__kernel void L3_SIMD_8x4_8x8_barrier(
    const __global float4 *src0,
    const __global float4 *src1,
    __global float4 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);
    // Result ctile is M rows x N columns
    // M = 8, we have 1 rows of work-items, so we need 8/1 = 8 results down
    // N = 32, we have 8 columns of work-items, so we need 32/8 = 4 results across = 1 float4s across

    float4 dot00 = (float4)(0.f);
    float4 dot01 = (float4)(0.f);
    float4 dot02 = (float4)(0.f);
    float4 dot03 = (float4)(0.f);
    float4 dot04 = (float4)(0.f);
    float4 dot05 = (float4)(0.f);
    float4 dot06 = (float4)(0.f);
    float4 dot07 = (float4)(0.f);

    __global float4 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( group_y * TILE_M + 8 * local_y) * width1;

    // Src0 is used directly as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    // M = 8, we have 1 rows of work-items, so we need 8/1 = 8 rows.
    // K = 32, we have 8 columns of work-items, so we need 32/8 = 4 floats across = 1 float4s across
    const __global float4 *src0_read = src0 + local_x + ( group_y * TILE_M + 8 * local_y) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    // K = 32, we'll process four rows at a time
    // N = 32, we have 8 columns of work-items, so we need 32/8 = 4 floats across = 1 float4s across
    const __global float4 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        const float4 arow0 = src0_read[ 0 * width0 ];
        const float4 arow1 = src0_read[ 1 * width0 ];
        const float4 arow2 = src0_read[ 2 * width0 ];
        const float4 arow3 = src0_read[ 3 * width0 ];
        const float4 arow4 = src0_read[ 4 * width0 ];
        const float4 arow5 = src0_read[ 5 * width0 ];
        const float4 arow6 = src0_read[ 6 * width0 ];
        const float4 arow7 = src0_read[ 7 * width0 ];

#define ITERATION( _index ) \
        {   \
            const float4 a0 = intel_sub_group_shuffle( arow0, _index ); \
            const float4 a1 = intel_sub_group_shuffle( arow1, _index ); \
            const float4 a2 = intel_sub_group_shuffle( arow2, _index ); \
            const float4 a3 = intel_sub_group_shuffle( arow3, _index ); \
            const float4 a4 = intel_sub_group_shuffle( arow4, _index ); \
            const float4 a5 = intel_sub_group_shuffle( arow5, _index ); \
            const float4 a6 = intel_sub_group_shuffle( arow6, _index ); \
            const float4 a7 = intel_sub_group_shuffle( arow7, _index ); \
            const float4 brow00 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow01 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow02 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow03 = src1_read0[ 0 ];   src1_read0 += width1;    \
			barrier(CLK_LOCAL_MEM_FENCE);	\
            dot00 = mad(brow00, (float4) a0.x, dot00);  \
            dot00 = mad(brow01, (float4) a0.y, dot00);  \
            dot00 = mad(brow02, (float4) a0.z, dot00);  \
            dot00 = mad(brow03, (float4) a0.w, dot00);  \
            dot01 = mad(brow00, (float4) a1.x, dot01);  \
            dot01 = mad(brow01, (float4) a1.y, dot01);  \
            dot01 = mad(brow02, (float4) a1.z, dot01);  \
            dot01 = mad(brow03, (float4) a1.w, dot01);  \
            dot02 = mad(brow00, (float4) a2.x, dot02);  \
            dot02 = mad(brow01, (float4) a2.y, dot02);  \
            dot02 = mad(brow02, (float4) a2.z, dot02);  \
            dot02 = mad(brow03, (float4) a2.w, dot02);  \
            dot03 = mad(brow00, (float4) a3.x, dot03);  \
            dot03 = mad(brow01, (float4) a3.y, dot03);  \
            dot03 = mad(brow02, (float4) a3.z, dot03);  \
            dot03 = mad(brow03, (float4) a3.w, dot03);  \
            dot04 = mad(brow00, (float4) a4.x, dot04);  \
            dot04 = mad(brow01, (float4) a4.y, dot04);  \
            dot04 = mad(brow02, (float4) a4.z, dot04);  \
            dot04 = mad(brow03, (float4) a4.w, dot04);  \
            dot05 = mad(brow00, (float4) a5.x, dot05);  \
            dot05 = mad(brow01, (float4) a5.y, dot05);  \
            dot05 = mad(brow02, (float4) a5.z, dot05);  \
            dot05 = mad(brow03, (float4) a5.w, dot05);  \
            dot06 = mad(brow00, (float4) a6.x, dot06);  \
            dot06 = mad(brow01, (float4) a6.y, dot06);  \
            dot06 = mad(brow02, (float4) a6.z, dot06);  \
            dot06 = mad(brow03, (float4) a6.w, dot06);  \
            dot07 = mad(brow00, (float4) a7.x, dot07);  \
            dot07 = mad(brow01, (float4) a7.y, dot07);  \
            dot07 = mad(brow02, (float4) a7.z, dot07);  \
            dot07 = mad(brow03, (float4) a7.w, dot07);  \
        }

        // We need K/VEC_SIZE iterations.
        // K = 32, VEC_SIZE = 4
        // So, 32/4 = 8 iterations.
        ITERATION( 0 );
        ITERATION( 1 );
        ITERATION( 2 );
        ITERATION( 3 );
        ITERATION( 4 );
        ITERATION( 5 );
        ITERATION( 6 );
        ITERATION( 7 );

#undef ITERATION

        src0_read += TILE_K / VEC_SIZE;
        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    dst_write0[ 0 ] = dot00;  dst_write0 += width1;
    dst_write0[ 0 ] = dot01;  dst_write0 += width1;
    dst_write0[ 0 ] = dot02;  dst_write0 += width1;
    dst_write0[ 0 ] = dot03;  dst_write0 += width1;
    dst_write0[ 0 ] = dot04;  dst_write0 += width1;
    dst_write0[ 0 ] = dot05;  dst_write0 += width1;
    dst_write0[ 0 ] = dot06;  dst_write0 += width1;
    dst_write0[ 0 ] = dot07;  dst_write0 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N

#define VEC_SIZE        4
#define TILE_M          64
#define TILE_K          32
#define TILE_N          32

__attribute__((reqd_work_group_size(8, 8, 1)))
__kernel void L3_SIMD_8x4_8x8(
    const __global float4 *src0,
    const __global float4 *src1,
    __global float4 *dst,const float alpha,
    const float beta,
    int width0,
    int width1)
{
    width0 /= VEC_SIZE;
    width1 /= VEC_SIZE;

    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);

    // Result ctile is M rows x N columns
    // M = 8, we have 1 rows of work-items, so we need 8/1 = 8 results down
    // N = 32, we have 8 columns of work-items, so we need 32/8 = 4 results across = 1 float4s across

    float4 dot00 = (float4)(0.f);
    float4 dot01 = (float4)(0.f);
    float4 dot02 = (float4)(0.f);
    float4 dot03 = (float4)(0.f);
    float4 dot04 = (float4)(0.f);
    float4 dot05 = (float4)(0.f);
    float4 dot06 = (float4)(0.f);
    float4 dot07 = (float4)(0.f);

    __global float4 *dst_write0 = dst + local_x + ( group_x * ( TILE_N / VEC_SIZE ) ) + ( group_y * TILE_M + 8 * local_y ) * width1;

    // Src0 is used directly as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    // M = 8, we have 1 rows of work-items, so we need 8/1 = 8 rows.
    // K = 32, we have 8 columns of work-items, so we need 32/8 = 4 floats across = 1 float4s across
    const __global float4 *src0_read = src0 + local_x + ( group_y * TILE_M + 8 * local_y ) * width0;

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    // K = 32, we'll process four rows at a time
    // N = 32, we have 8 columns of work-items, so we need 32/8 = 4 floats across = 1 float4s across
    const __global float4 *src1_read0 = src1 + local_x + ( group_x * ( TILE_N / VEC_SIZE ) );

    // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        const float4 arow0 = src0_read[ 0 * width0 ];
        const float4 arow1 = src0_read[ 1 * width0 ];
        const float4 arow2 = src0_read[ 2 * width0 ];
        const float4 arow3 = src0_read[ 3 * width0 ];
        const float4 arow4 = src0_read[ 4 * width0 ];
        const float4 arow5 = src0_read[ 5 * width0 ];
        const float4 arow6 = src0_read[ 6 * width0 ];
        const float4 arow7 = src0_read[ 7 * width0 ];

#define ITERATION( _index ) \
        {   \
            const float4 a0 = intel_sub_group_shuffle( arow0, _index ); \
            const float4 a1 = intel_sub_group_shuffle( arow1, _index ); \
            const float4 a2 = intel_sub_group_shuffle( arow2, _index ); \
            const float4 a3 = intel_sub_group_shuffle( arow3, _index ); \
            const float4 a4 = intel_sub_group_shuffle( arow4, _index ); \
            const float4 a5 = intel_sub_group_shuffle( arow5, _index ); \
            const float4 a6 = intel_sub_group_shuffle( arow6, _index ); \
            const float4 a7 = intel_sub_group_shuffle( arow7, _index ); \
            const float4 brow00 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow01 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow02 = src1_read0[ 0 ];   src1_read0 += width1;    \
            const float4 brow03 = src1_read0[ 0 ];   src1_read0 += width1;    \
            dot00 = mad(brow00, (float4) a0.x, dot00);  \
            dot00 = mad(brow01, (float4) a0.y, dot00);  \
            dot00 = mad(brow02, (float4) a0.z, dot00);  \
            dot00 = mad(brow03, (float4) a0.w, dot00);  \
            dot01 = mad(brow00, (float4) a1.x, dot01);  \
            dot01 = mad(brow01, (float4) a1.y, dot01);  \
            dot01 = mad(brow02, (float4) a1.z, dot01);  \
            dot01 = mad(brow03, (float4) a1.w, dot01);  \
            dot02 = mad(brow00, (float4) a2.x, dot02);  \
            dot02 = mad(brow01, (float4) a2.y, dot02);  \
            dot02 = mad(brow02, (float4) a2.z, dot02);  \
            dot02 = mad(brow03, (float4) a2.w, dot02);  \
            dot03 = mad(brow00, (float4) a3.x, dot03);  \
            dot03 = mad(brow01, (float4) a3.y, dot03);  \
            dot03 = mad(brow02, (float4) a3.z, dot03);  \
            dot03 = mad(brow03, (float4) a3.w, dot03);  \
            dot04 = mad(brow00, (float4) a4.x, dot04);  \
            dot04 = mad(brow01, (float4) a4.y, dot04);  \
            dot04 = mad(brow02, (float4) a4.z, dot04);  \
            dot04 = mad(brow03, (float4) a4.w, dot04);  \
            dot05 = mad(brow00, (float4) a5.x, dot05);  \
            dot05 = mad(brow01, (float4) a5.y, dot05);  \
            dot05 = mad(brow02, (float4) a5.z, dot05);  \
            dot05 = mad(brow03, (float4) a5.w, dot05);  \
            dot06 = mad(brow00, (float4) a6.x, dot06);  \
            dot06 = mad(brow01, (float4) a6.y, dot06);  \
            dot06 = mad(brow02, (float4) a6.z, dot06);  \
            dot06 = mad(brow03, (float4) a6.w, dot06);  \
            dot07 = mad(brow00, (float4) a7.x, dot07);  \
            dot07 = mad(brow01, (float4) a7.y, dot07);  \
            dot07 = mad(brow02, (float4) a7.z, dot07);  \
            dot07 = mad(brow03, (float4) a7.w, dot07);  \
        }

        // We need K/VEC_SIZE iterations.
        // K = 32, VEC_SIZE = 4
        // So, 32/4 = 8 iterations.
        ITERATION( 0 );
        ITERATION( 1 );
        ITERATION( 2 );
        ITERATION( 3 );
        ITERATION( 4 );
        ITERATION( 5 );
        ITERATION( 6 );
        ITERATION( 7 );

#undef ITERATION

        src0_read += TILE_K / VEC_SIZE;
        w += TILE_K / VEC_SIZE;
    }
    while( w < width0 );

    dst_write0[ 0 ] = dot00;  dst_write0 += width1;
    dst_write0[ 0 ] = dot01;  dst_write0 += width1;
    dst_write0[ 0 ] = dot02;  dst_write0 += width1;
    dst_write0[ 0 ] = dot03;  dst_write0 += width1;
    dst_write0[ 0 ] = dot04;  dst_write0 += width1;
    dst_write0[ 0 ] = dot05;  dst_write0 += width1;
    dst_write0[ 0 ] = dot06;  dst_write0 += width1;
    dst_write0[ 0 ] = dot07;  dst_write0 += width1;
}

#undef VEC_SIZE
#undef TILE_M
#undef TILE_K
#undef TILE_N


#define TILE_M          4
#define TILE_K          8
#define TILE_N          8

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void block_read_4x1_1x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
    const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 4, we have 1 row of work-items, so we need 4/1 = 4 results down
    // N = 8, we have 8 columns of work-items, so we need 8/8 = 1 result across

    float4  blockC00 = 0.f;

    __global float *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M ) * width1;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M );

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

	 // Walk ACROSS src0 and DOWN src1:
    int w = 0;
    do
    {
        // We want to load atile, which is M rows x K columns
        // M = 4, we have 1 row of work-items, so each work-item must load 4/1 = 4 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        float4  blockA = as_float4( intel_sub_group_block_read4( src0, coordA ) );   coordA.x += TILE_K * sizeof(uint);

        // Now load btile, which is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        float8  blockB = as_float8( intel_sub_group_block_read8( src1, coordB ) );   coordB.y += TILE_K;

#define TRANSPOSE_BLOCK_4( _block, _col )   \
        (float4)( intel_sub_group_shuffle( _block, _col ));

#define MULTIPLY_BLOCKS_4x8( _result, _blockA, _blockB )    \
        {   \
            const float4    acol0 = TRANSPOSE_BLOCK_4( _blockA, 0 );    \
            const float4    acol1 = TRANSPOSE_BLOCK_4( _blockA, 1 );    \
            const float4    acol2 = TRANSPOSE_BLOCK_4( _blockA, 2 );    \
            const float4    acol3 = TRANSPOSE_BLOCK_4( _blockA, 3 );    \
            const float4    acol4 = TRANSPOSE_BLOCK_4( _blockA, 4 );    \
            const float4    acol5 = TRANSPOSE_BLOCK_4( _blockA, 5 );    \
            const float4    acol6 = TRANSPOSE_BLOCK_4( _blockA, 6 );    \
            const float4    acol7 = TRANSPOSE_BLOCK_4( _blockA, 7 );    \
            _result = mad( (float4)(_blockB.s0), acol0, _result );    \
            _result = mad( (float4)(_blockB.s1), acol1, _result );    \
            _result = mad( (float4)(_blockB.s2), acol2, _result );    \
            _result = mad( (float4)(_blockB.s3), acol3, _result );    \
            _result = mad( (float4)(_blockB.s4), acol4, _result );    \
            _result = mad( (float4)(_blockB.s5), acol5, _result );    \
            _result = mad( (float4)(_blockB.s6), acol6, _result );    \
            _result = mad( (float4)(_blockB.s7), acol7, _result );    \
        }

        MULTIPLY_BLOCKS_4x8( blockC00, blockA, blockB );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_4x8

        w += TILE_K;
    }
    while( w < width0 );

    dst_write0[0] = blockC00.s0;   dst_write0 += width1;
    dst_write0[0] = blockC00.s1;   dst_write0 += width1;
    dst_write0[0] = blockC00.s2;   dst_write0 += width1;
    dst_write0[0] = blockC00.s3;   dst_write0 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N


////////////////////////////////////////////////////////////////
// MediaBlockRead_SIMD_2x32

#define TILE_M          32
#define TILE_K          8
#define TILE_N          16

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void block_read_32x2_1x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC00 = 0.0f;
    float8 blockC01 = 0.0f;
    float8 blockC02 = 0.0f;
    float8 blockC03 = 0.0f;
    float8 blockC10 = 0.0f;
    float8 blockC11 = 0.0f;
    float8 blockC12 = 0.0f;
    float8 blockC13 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M );

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block.s0, _col ),   \
                  intel_sub_group_shuffle( _block.s1, _col ),   \
                  intel_sub_group_shuffle( _block.s2, _col ),   \
                  intel_sub_group_shuffle( _block.s3, _col ),   \
                  intel_sub_group_shuffle( _block.s4, _col ),   \
                  intel_sub_group_shuffle( _block.s5, _col ),   \
                  intel_sub_group_shuffle( _block.s6, _col ),   \
                  intel_sub_group_shuffle( _block.s7, _col ) );

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC00, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC10, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC01, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC11, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC02, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC12, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC03, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC13, blockA, blockB10 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M ) * width1;
    __global float  *dst_write1 = dst_write0 + ( TILE_N / 2 );

    dst_write0[ 0 ] = blockC00.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC10.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s7; dst_write1 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N


#define TILE_M          128
#define TILE_K          8
#define TILE_N          16

__attribute__((reqd_work_group_size(8, 4, 1)))
__kernel void block_read_32x2_4x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);
    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC00 = 0.0f;
    float8 blockC01 = 0.0f;
    float8 blockC02 = 0.0f;
    float8 blockC03 = 0.0f;
    float8 blockC10 = 0.0f;
    float8 blockC11 = 0.0f;
    float8 blockC12 = 0.0f;
    float8 blockC13 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M + 32*local_y);

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block.s0, _col ),   \
                  intel_sub_group_shuffle( _block.s1, _col ),   \
                  intel_sub_group_shuffle( _block.s2, _col ),   \
                  intel_sub_group_shuffle( _block.s3, _col ),   \
                  intel_sub_group_shuffle( _block.s4, _col ),   \
                  intel_sub_group_shuffle( _block.s5, _col ),   \
                  intel_sub_group_shuffle( _block.s6, _col ),   \
                  intel_sub_group_shuffle( _block.s7, _col ) );

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC00, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC10, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC01, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC11, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC02, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC12, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC03, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC13, blockA, blockB10 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M  + 32*local_y) * width1;
    __global float  *dst_write1 = dst_write0 + ( TILE_N / 2 );

    dst_write0[ 0 ] = blockC00.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC10.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s7; dst_write1 += width1;
}
#undef TILE_M
#undef TILE_K
#undef TILE_N

#define TILE_M          256
#define TILE_K          8
#define TILE_N          16

__attribute__((reqd_work_group_size(8, 8, 1)))
__kernel void block_read_32x2_8x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);
    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC00 = 0.0f;
    float8 blockC01 = 0.0f;
    float8 blockC02 = 0.0f;
    float8 blockC03 = 0.0f;
    float8 blockC10 = 0.0f;
    float8 blockC11 = 0.0f;
    float8 blockC12 = 0.0f;
    float8 blockC13 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M + 32*local_y);

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block.s0, _col ),   \
                  intel_sub_group_shuffle( _block.s1, _col ),   \
                  intel_sub_group_shuffle( _block.s2, _col ),   \
                  intel_sub_group_shuffle( _block.s3, _col ),   \
                  intel_sub_group_shuffle( _block.s4, _col ),   \
                  intel_sub_group_shuffle( _block.s5, _col ),   \
                  intel_sub_group_shuffle( _block.s6, _col ),   \
                  intel_sub_group_shuffle( _block.s7, _col ) );

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC00, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC10, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC01, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC11, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC02, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC12, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC03, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC13, blockA, blockB10 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M  + 32*local_y) * width1;
    __global float  *dst_write1 = dst_write0 + ( TILE_N / 2 );

    dst_write0[ 0 ] = blockC00.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC10.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC12.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC13.s7; dst_write1 += width1;
}
#undef TILE_M
#undef TILE_K
#undef TILE_N

#define TILE_M          32
#define TILE_K          8
#define TILE_N          8

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void block_read_32x1_1x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 8, we have 8 columns of work-items, so we need 8/8 = 1 results across

    float8 blockC00 = 0.0f;
    float8 blockC01 = 0.0f;
    float8 blockC02 = 0.0f;
    float8 blockC03 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M );

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );
    // Walk ACROSS src0 and DOWN src1:

    int w = 0;
    do
    {

#define TRANSPOSE_BLOCK_8( _block, _col )   \
        ( intel_sub_group_shuffle( _block, _col ));

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );    \
            _result = mad( (float8)(_blockB.s1), acol1, _result );    \
            _result = mad( (float8)(_blockB.s2), acol2, _result );    \
            _result = mad( (float8)(_blockB.s3), acol3, _result );    \
            _result = mad( (float8)(_blockB.s4), acol4, _result );    \
            _result = mad( (float8)(_blockB.s5), acol5, _result );    \
            _result = mad( (float8)(_blockB.s6), acol6, _result );    \
            _result = mad( (float8)(_blockB.s7), acol7, _result );    \
        }

        // Now load btile, which is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );

        coordB.y += TILE_K;

		// We want to load atile, which is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );   coordATemp.y += 8;
		MULTIPLY_BLOCKS_8x8( blockC00, blockA, blockB00 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );   coordATemp.y += 8;
		MULTIPLY_BLOCKS_8x8( blockC01, blockA, blockB00 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );   coordATemp.y += 8;
		MULTIPLY_BLOCKS_8x8( blockC02, blockA, blockB00 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );   coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC03, blockA, blockB00 );

        coordA.x += TILE_K * sizeof(uint);

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8

        w += TILE_K;
    }
    while( w < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M ) * width1;

    dst_write0[ 0 ] = blockC00.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC02.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC03.s7; dst_write0 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N

#define TILE_M          16
#define TILE_K          8
#define TILE_N          16

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void block_read_16x2_1x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC00 = 0.0f;
    float8 blockC01 = 0.0f;
    float8 blockC10 = 0.0f;
    float8 blockC11 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M );

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block.s0, _col ),   \
                  intel_sub_group_shuffle( _block.s1, _col ),   \
                  intel_sub_group_shuffle( _block.s2, _col ),   \
                  intel_sub_group_shuffle( _block.s3, _col ),   \
                  intel_sub_group_shuffle( _block.s4, _col ),   \
                  intel_sub_group_shuffle( _block.s5, _col ),   \
                  intel_sub_group_shuffle( _block.s6, _col ),   \
                  intel_sub_group_shuffle( _block.s7, _col ) );

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC00, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC10, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC01, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC11, blockA, blockB10 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M ) * width1;
    __global float  *dst_write1 = dst_write0 + ( TILE_N / 2 );

    dst_write0[ 0 ] = blockC00.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC10.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s7; dst_write1 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N


#define TILE_M          64
#define TILE_K          8
#define TILE_N          16

__attribute__((reqd_work_group_size(8, 4, 1)))
__kernel void block_read_16x2_4x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);
    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC00 = 0.0f;
    float8 blockC01 = 0.0f;
    float8 blockC10 = 0.0f;
    float8 blockC11 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M + local_y * 16);

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block.s0, _col ),   \
                  intel_sub_group_shuffle( _block.s1, _col ),   \
                  intel_sub_group_shuffle( _block.s2, _col ),   \
                  intel_sub_group_shuffle( _block.s3, _col ),   \
                  intel_sub_group_shuffle( _block.s4, _col ),   \
                  intel_sub_group_shuffle( _block.s5, _col ),   \
                  intel_sub_group_shuffle( _block.s6, _col ),   \
                  intel_sub_group_shuffle( _block.s7, _col ) );

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC00, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC10, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC01, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC11, blockA, blockB10 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M + local_y * 16) * width1;
    __global float  *dst_write1 = dst_write0 + ( TILE_N / 2 );

    dst_write0[ 0 ] = blockC00.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC10.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s7; dst_write1 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N


#define TILE_M          128
#define TILE_K          8
#define TILE_N          16

__attribute__((reqd_work_group_size(8, 8, 1)))
__kernel void block_read_16x2_8x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);
    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC00 = 0.0f;
    float8 blockC01 = 0.0f;
    float8 blockC10 = 0.0f;
    float8 blockC11 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M + local_y * 16);

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block.s0, _col ),   \
                  intel_sub_group_shuffle( _block.s1, _col ),   \
                  intel_sub_group_shuffle( _block.s2, _col ),   \
                  intel_sub_group_shuffle( _block.s3, _col ),   \
                  intel_sub_group_shuffle( _block.s4, _col ),   \
                  intel_sub_group_shuffle( _block.s5, _col ),   \
                  intel_sub_group_shuffle( _block.s6, _col ),   \
                  intel_sub_group_shuffle( _block.s7, _col ) );

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC00, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC10, blockA, blockB10 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC01, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC11, blockA, blockB10 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M + local_y * 16) * width1;
    __global float  *dst_write1 = dst_write0 + ( TILE_N / 2 );

    dst_write0[ 0 ] = blockC00.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC00.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC01.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC10.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC10.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC11.s7; dst_write1 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N


#define TILE_M          16
#define TILE_K          8
#define TILE_N          32

__attribute__((reqd_work_group_size(8, 1, 1)))
__kernel void block_read_16x4_1x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);

    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC0 = 0.0f;
    float8 blockC1 = 0.0f;
	float8 blockC2 = 0.0f;
    float8 blockC3 = 0.0f;

    float8 blockC4 = 0.0f;
    float8 blockC5 = 0.0f;
	float8 blockC6 = 0.0f;
    float8 blockC7 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M );

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block, _col ));

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
		float8  blockB20 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB30 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC0, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC2, blockA, blockB10 );
		MULTIPLY_BLOCKS_8x8( blockC4, blockA, blockB20 );
        MULTIPLY_BLOCKS_8x8( blockC6, blockA, blockB30 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC1, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC3, blockA, blockB10 );
        MULTIPLY_BLOCKS_8x8( blockC5, blockA, blockB20 );
        MULTIPLY_BLOCKS_8x8( blockC7, blockA, blockB30 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M ) * width1;
    __global float  *dst_write1 = dst_write0 + ( 8 );
	__global float  *dst_write2 = dst_write1 + ( 8 );
	__global float  *dst_write3 = dst_write2 + ( 8 );

    dst_write0[ 0 ] = blockC0.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC2.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s7; dst_write1 += width1;

	dst_write2[ 0 ] = blockC4.s0; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s1; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s2; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s3; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s4; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s5; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s6; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s7; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s0; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s1; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s2; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s3; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s4; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s5; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s6; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s7; dst_write2 += width1;

    dst_write3[ 0 ] = blockC6.s0; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s1; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s2; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s3; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s4; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s5; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s6; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s7; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s0; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s1; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s2; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s3; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s4; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s5; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s6; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s7; dst_write3 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N


#define TILE_M          64
#define TILE_K          8
#define TILE_N          32

__attribute__((reqd_work_group_size(8, 4, 1)))
__kernel void block_read_16x4_4x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);
    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC0 = 0.0f;
    float8 blockC1 = 0.0f;
	float8 blockC2 = 0.0f;
    float8 blockC3 = 0.0f;

    float8 blockC4 = 0.0f;
    float8 blockC5 = 0.0f;
	float8 blockC6 = 0.0f;
    float8 blockC7 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M + local_y * 16);

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block, _col ));

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
		float8  blockB20 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB30 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC0, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC2, blockA, blockB10 );
		MULTIPLY_BLOCKS_8x8( blockC4, blockA, blockB20 );
        MULTIPLY_BLOCKS_8x8( blockC6, blockA, blockB30 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC1, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC3, blockA, blockB10 );
        MULTIPLY_BLOCKS_8x8( blockC5, blockA, blockB20 );
        MULTIPLY_BLOCKS_8x8( blockC7, blockA, blockB30 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M + local_y * 16) * width1;
    __global float  *dst_write1 = dst_write0 + ( 8 );
	__global float  *dst_write2 = dst_write1 + ( 8 );
	__global float  *dst_write3 = dst_write2 + ( 8 );

    dst_write0[ 0 ] = blockC0.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC2.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s7; dst_write1 += width1;

	dst_write2[ 0 ] = blockC4.s0; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s1; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s2; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s3; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s4; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s5; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s6; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s7; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s0; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s1; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s2; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s3; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s4; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s5; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s6; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s7; dst_write2 += width1;

    dst_write3[ 0 ] = blockC6.s0; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s1; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s2; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s3; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s4; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s5; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s6; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s7; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s0; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s1; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s2; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s3; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s4; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s5; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s6; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s7; dst_write3 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N


#define TILE_M          128
#define TILE_K          8
#define TILE_N          32

__attribute__((reqd_work_group_size(8, 8, 1)))
__kernel void block_read_16x4_8x8(
    __read_only image2d_t src0,
    __read_only image2d_t src1,
    __global float *dst,
	const float alpha,
    const float beta,
    int width0,
    int width1)
{
    const int group_x = get_group_id(0);
    const int group_y = get_group_id(1);
    const int local_x = get_local_id(0);
	const int local_y = get_local_id(1);
    // Result ctile is M rows x N columns
    // M = 32, we have 1 rows of work-items, so we need 32/1 = 32 results down = 4 x float8
    // N = 16, we have 8 columns of work-items, so we need 16/8 = 2 results across
    // Note: ctile is 64 registers

    float8 blockC0 = 0.0f;
    float8 blockC1 = 0.0f;
	float8 blockC2 = 0.0f;
    float8 blockC3 = 0.0f;

    float8 blockC4 = 0.0f;
    float8 blockC5 = 0.0f;
	float8 blockC6 = 0.0f;
    float8 blockC7 = 0.0f;

    // Src0 is directly used as atile.
    // It starts at the left side of src0 and walks across.
    // atile is M rows x K columns.
    int2    coordA = (int2)( 0, group_y * TILE_M + local_y * 16);

    // Src1 is directly used as btile.
    // It starts at the top of src1 and walks down.
    // btile is K rows x N columns.
    int2    coordB = (int2)( ( group_x * TILE_N ) * sizeof(uint), 0 );

    // Walk ACROSS src0 and DOWN src1:
    do
    {
#define TRANSPOSE_BLOCK_8( _block, _col )   \
        (float8)( intel_sub_group_shuffle( _block, _col ));

#define MULTIPLY_BLOCKS_8x8( _result, _blockA, _blockB )    \
        {   \
            const float8    acol0 = TRANSPOSE_BLOCK_8( _blockA, 0 );    \
            const float8    acol1 = TRANSPOSE_BLOCK_8( _blockA, 1 );    \
            const float8    acol2 = TRANSPOSE_BLOCK_8( _blockA, 2 );    \
            const float8    acol3 = TRANSPOSE_BLOCK_8( _blockA, 3 );    \
            const float8    acol4 = TRANSPOSE_BLOCK_8( _blockA, 4 );    \
            const float8    acol5 = TRANSPOSE_BLOCK_8( _blockA, 5 );    \
            const float8    acol6 = TRANSPOSE_BLOCK_8( _blockA, 6 );    \
            const float8    acol7 = TRANSPOSE_BLOCK_8( _blockA, 7 );    \
            _result = mad( (float8)(_blockB.s0), acol0, _result );      \
            _result = mad( (float8)(_blockB.s1), acol1, _result );      \
            _result = mad( (float8)(_blockB.s2), acol2, _result );      \
            _result = mad( (float8)(_blockB.s3), acol3, _result );      \
            _result = mad( (float8)(_blockB.s4), acol4, _result );      \
            _result = mad( (float8)(_blockB.s5), acol5, _result );      \
            _result = mad( (float8)(_blockB.s6), acol6, _result );      \
            _result = mad( (float8)(_blockB.s7), acol7, _result );      \
        }

        // atile is M rows x K columns
        // M = 32, we have 1 row of work-items, so each work-item must load 32/1 = 32 rows
        // K = 8, we have 8 columns of work-items, so each work-item must load 8/8 = 1 column
        // Note: atile is up to 32 registers

        // btile is K rows x N columns
        // K = 8, we have 1 row of work-items, so each work-item must load 8/1 = 8 rows
        // N = 16, we have 8 columns of work-items, so each work-item must load 16/8 = 2 column
        // Note: btile is up to 16 registers

        int2    coordBTemp = coordB;
        float8  blockB00 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB10 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
		float8  blockB20 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordBTemp.x += 8 * sizeof(uint);
        float8  blockB30 = as_float8( intel_sub_group_block_read8( src1, coordBTemp ) );    coordB.y += TILE_K;

        int2    coordATemp = coordA;
        float8  blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordATemp.y += 8;
        MULTIPLY_BLOCKS_8x8( blockC0, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC2, blockA, blockB10 );
		MULTIPLY_BLOCKS_8x8( blockC4, blockA, blockB20 );
        MULTIPLY_BLOCKS_8x8( blockC6, blockA, blockB30 );

        blockA = as_float8( intel_sub_group_block_read8( src0, coordATemp ) );    coordA.x += TILE_K * sizeof(uint);
        MULTIPLY_BLOCKS_8x8( blockC1, blockA, blockB00 );
        MULTIPLY_BLOCKS_8x8( blockC3, blockA, blockB10 );
        MULTIPLY_BLOCKS_8x8( blockC5, blockA, blockB20 );
        MULTIPLY_BLOCKS_8x8( blockC7, blockA, blockB30 );

#undef TRANSPOSE_BLOCK_8
#undef MULTIPLY_BLOCKS_8x8
    }
    while( coordB.y < width0 );

    __global float  *dst_write0 = dst + local_x + ( group_x * ( TILE_N ) ) + ( group_y * TILE_M + local_y * 16) * width1;
    __global float  *dst_write1 = dst_write0 + ( 8 );
	__global float  *dst_write2 = dst_write1 + ( 8 );
	__global float  *dst_write3 = dst_write2 + ( 8 );

    dst_write0[ 0 ] = blockC0.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC0.s7; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s0; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s1; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s2; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s3; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s4; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s5; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s6; dst_write0 += width1;
    dst_write0[ 0 ] = blockC1.s7; dst_write0 += width1;

    dst_write1[ 0 ] = blockC2.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC2.s7; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s0; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s1; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s2; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s3; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s4; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s5; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s6; dst_write1 += width1;
    dst_write1[ 0 ] = blockC3.s7; dst_write1 += width1;

	dst_write2[ 0 ] = blockC4.s0; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s1; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s2; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s3; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s4; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s5; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s6; dst_write2 += width1;
    dst_write2[ 0 ] = blockC4.s7; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s0; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s1; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s2; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s3; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s4; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s5; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s6; dst_write2 += width1;
    dst_write2[ 0 ] = blockC5.s7; dst_write2 += width1;

    dst_write3[ 0 ] = blockC6.s0; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s1; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s2; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s3; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s4; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s5; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s6; dst_write3 += width1;
    dst_write3[ 0 ] = blockC6.s7; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s0; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s1; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s2; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s3; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s4; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s5; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s6; dst_write3 += width1;
    dst_write3[ 0 ] = blockC7.s7; dst_write3 += width1;
}

#undef TILE_M
#undef TILE_K
#undef TILE_N

#endif
